function [ x t ] = WTSS_EMP_Scheme( x )
%
% This function creates the WTSS-EMP scheme based on wavelets and EMP.
%
% Parameters:
%    @INPUT Hyperspectral image 'x'
%
%    @OUTPUT 'x' WTSS-EMP
%            't' Execution time
%
% Example:
% [ x t ] = WTSS_EMP_Scheme( x );
% 
% Output obtained:
%
%  Name        Size                     Bytes  Class
%  t           1x5                         40  double              
%  x         610x340x139            230628800  double
%
% t =
%    1.1694    0.8362    2.7181    7.7053    0.3392
% x =
%    This is the hyperspectral data produced by the WTSS-EMP scheme
 

% Selesnik framework is available at http://eeweb.poly.edu/iselesni/DoubleSoftware/ 
% WAVELET SOFTWARE AT POLYTECHNIC UNIVERSITY, BROOKLYN, NY.
addpath './Double-Density Wavelet Software/allcode/'

% t(1,:) has the time for FR, EMP, 1D NeighShrink, and 2D NeighShrink, Format data
t = zeros(1,5);

      
    % size of the image (rows cols bands)
    size_x = size(x);

    % t_format_data is used in several places
    t_format_data = 0;
    
    % number of levels for 1D wavelet denoising
    L = 1;

    tic;
        % Feature reduction by wavelets
        HSI_1DWT = cdf97_forward(row_major(x), L);
    t_forwardDWT1d = toc;
    
    tic;
    % first L components are untouched
    HSI_FIRST = HSI_1DWT(:,1:ceil(size_x(3)/2^L));
    % and the rest of band coefficients are denoised similar in spectral domain
    HSI_REST  = HSI_1DWT(:,ceil(size_x(3)/2^L)+1:size(HSI_1DWT,2));
    t_format_data = t_format_data + toc;
    
    tic;
        % number of stages required for get at least 4 band-coefficients
        % depends on the number of spectral bands. The maximum level of
        % wavelet stages is log2(num_bands). So, if we do two less stages
        % from the maximum we get at least 4 band-coefficients
        k = ceil (log2(size_x(3))) - 2;
        % for the emp keep reducing 
        HSI_FE = cdf97_forward(HSI_FIRST, k-L);
        % and get 4 band coefficients
        HSI_FE = HSI_FE(:,1:4);
    t_FR = toc;
    
 
    tic;
        % prepare data for EMP
        HSI_FE = row_major_restore( HSI_FE, [ size_x(1) size_x(2) size(HSI_FE,2) ] );
        % data should be normalized
        HSI_FE = scale_data(HSI_FE, 0, 255);
        % to be exactly as in the CPU and GPU versions
        HSI_FE = floor(HSI_FE + 0.5);
        % and change the data type
        HSI_FE = uint8(HSI_FE);
    t_format_data = t_format_data + toc;
    
    tic;
        % Extended Morphological Profile
        HSI_EMP = emp( HSI_FE, size(HSI_FE, 3) );
    t_emp = toc;
    
    tic;
        HSI_EMP = double(HSI_EMP);
        % data should be normalized
        HSI_EMP = scale_data(HSI_EMP, 0, 1);
    t_format_data = t_format_data + toc;
        
    %% Denoising the spectral domain of the H coefficients
    tic;
        % wavelet coefficients at the finest first level HH1
        % Nos quedamos con la primera mitad del vector correspondiente a H
        % después de aplicar una vez la wavelet 1D. Como le hemos quitado
        % el primer trozo, le tenemos que quitar ceil(size_x(3)/2^L)
        H = HSI_REST(:,1:ceil(size_x(3)/2^L));
        % Noise standard deviation
        sigma = median(abs(H(:)))/0.6745;
        % number of pixel of the image
        N = numel(H);
        % Universal threshold
        T = sigma*sqrt(2*log(N));
                
        HSI_SHRINK = neighShrink1d(HSI_REST, T);
    
    t_neighShrink1d = toc;
    
    tic;
    % prepare the image with the first L components and the shrinkage
    HSI_DENOISED = [ HSI_FIRST HSI_SHRINK ];
    t_format_data = t_format_data + toc;

    clear H HSI_FIRST HSI_REST HSI_SHRINK HSI_FE;
           
    tic;    
        % inverse DWT with the signal denoised in the spectral domain
        HSI_DENOISED = cdf97_inverse(HSI_DENOISED, L);
        % Al aplicar la wavelet 1D puede que el numero de bandas aumente para
        % hacerlas pares y divisibles, por eso en HSI_REST solo almaceno hasta
        % size_x(3) que son las bandas originales
        HSI_DENOISED = HSI_DENOISED(:,1:size_x(3));
        % prepare as a cube for stack vector
        HSI_DENOISED = row_major_restore(HSI_DENOISED, [ size_x(1) size_x(2) size_x(3) ] );  
    
    t_inverseDWT1d = toc;

    tic;
        % Denoising the spatial domain of the restored image
        HSI_DENOISED = dwtDenoising_2DensityNeighShrink(HSI_DENOISED);
    t_denoising_2D_NeighShrink = toc;
    
    tic;
    % stack denoised image and EMP
    x = stack_vector( HSI_DENOISED, HSI_EMP);
    t_format_data = t_format_data + toc;


    % Feature Reduction
    t(1,1) = t_forwardDWT1d + t_FR;
    % Extended Morphological Profile
    t(1,2) = t_emp;
    % Spectral denoising
    t(1,3) = t_neighShrink1d + t_inverseDWT1d;
    % Spatial denoising
    t(1,4) = t_denoising_2D_NeighShrink;
    % time for formating data
    t(1,5) = t_format_data;
    
% diary('RESULTS_TIMES_WTSS_EMP.log');
%     t
% diary off;

end

function [ y ] = neighShrink1d ( w, t )
%% Neighboring threshold

windowsize  = 3;
windowfilt = ones(1,windowsize);

sij = conv2(1,windowfilt,w.^2,'same');
% The '+' sign at the end of the formula means to keep the value if it is
% posivite, and set it to zero otherwise.
y = w .* max(0, (1 - (t^2 ./ sij)));

end

