function [ w ] = idwt97_1D( a )
%
% Inverse 1D wavelet 9 / 7 applied to each row of the matrix a
%
% Authors: Francisco Arguello (C version) Pablo (MATLAB version)

% Once we arrive at our discrete wavelet coefficients, we need a way to 
% reconstruct them back into the original signal (or a modified original 
% signal if we played around with the coefficients). 
%
% In order to do this, we utilize the process known as the inverse discrete 
% wavelet transform.
%
% Much like the DWT can be explained by using filter bank theory, so can the 
% reconstruction of the IDWT.
%
% The process is simply reversed. The DWT coefficients are first upsampled 
% (the approximation and the detail coefficients are handled separately) 
% by placing zeros in between every coefficient, effectively doubling the 
% lengths of each. These are then convolved with the reconstruction scaling 
% filter for approximation coefficients (the reconstruction scaling filter 
% is simply the original scaling filter that has been flipped left to right)
% and the reconstruction wavelet filter for the detail coefficients. 
% These results are then added together to arrive at the original signal.  
    
M1 = 9;
M2 = 7;

hh  = [-0.0378284555,-0.0238494650,0.1106244044,0.3774028556,-0.8526986790,0.3774028556,0.1106244044,-0.0238494650,-0.0378284555];
gg = [-0.0645388826,-0.0406894176,0.4180922732,0.7884856164,0.4180922732,-0.0406894176,-0.0645388826];

% http://www.mathworks.com/matlabcentral/fileexchange/11846-cdf-97-wavelet-transform/content/wavecdf97.m
%% ---------------- forming low-pass and high-pass filters --------------- 
% CDF 9/7 filters: decomposition low-pass lp and high-pass hp
%                  reconstruction low-pass lpr and high-pass hpr
% The filter coefficients have several forms.
% What D.S.Taubman et al. suggest are used here:
% lp = [.026748757411 -.016864118443 -.078223266529 .266864118443];
% lp = [lp .602949018236 fliplr(lp)];
% hp = [.045635881557 -.028771763114 -.295635881557];
% hp = [hp .557543526229 fliplr(hp)];
% hh = hp .* [-1 1 -1 1 -1 1 -1] * 2;
% gg = lp .* [1 -1 1 -1 1 -1 1 -1 1] * 2;

% for readability
N  = size(a,2);


%% The DWT coefficients are first upsampled (the approximation and the detail 
% coefficients are handled separately) by placing zeros in between every coefficient,
% effectively doubling the lengths of each. 
% aa = upsample( a(:,1:ceil(N/2))  , 2 );
aa = upsamplecol( a, 1, ceil(N/2) );
% dd = upsample( a(:,ceil(N/2)+1:N), 2 );
dd = upsamplecol( a, ceil(N/2)+1, N );
%% pad input array so the convolution filter fit within the array
% Similar to how we made the signal periodic before doing our DWT calculations on it, 
% we must make our dwt coefficients periodic before convolving to obtain the original signal. 
% This is done by simply taking the first N/2-1 coefficients from the DWT coefficients,
% and appending them to the end. Remember that N is the length of our scaling filter.
% aa = padarray(aa, [0 9],  'post');
% dd = padarray(dd, [0 9], 'post');

aa = padarray(aa, [0 9], 'circular', 'post');
dd = padarray(dd, [0 9], 'circular', 'post');

% MATLAB tip: The size of the indicated variable or array appears to be 
% changing with each loop iteration. 
w = zeros(size(a,1),N);

% After the convolution and addition, to grab the part of the signal we want 
% away from the convolution ‘junk’, we grab the coefficients from N to the 
% length of the signal + N -1. This will give us our original signal.
il = numel(gg);    % we do it on the fly. Just write where corresponding!

    % for each coefficient in the spectral domain...
    for i=1:N
        lp=0;
        hp=0;
        
        for j=1:M2
            lp = lp + gg(j) * aa(:,i+j-1);  % low pass
        end
        for j=1:M1
            hp = hp + hh(j) * dd(:, i+j-1); % high pass
        end
        % update all the points with the corresponding value
        % we put the low pass in the first half of the array...
        w(:,il)   = lp + hp;
        il = il + 1;
        if (il > N) 
            il = 1;
        end;

    end

% After the convolution and addition, to grab the part of the signal we want 
% away from the convolution ‘junk’, we grab the coefficients from N to the 
% length of the signal + N -1. This will give us our original signal.

end

%% Implements a simple zero-padding scheme very useful in the wavelet
% reconstruction algorithm (dyadicup is within the Wavelet Toolbox). 
% This functions pad the matrix by columns because the 'upsample' function
% does the upsampling only by rows
%
% INPUT: x the matrix to upsample
%        a first element to upsample
%        b last element to upsample
% OUTPUT: y the upsample matrix by columns
function [ y ] = upsamplecol( x, a, b )
    y = x(:,a:b);
    y = transpose(y);
    y = upsample(y, 2);
    y = transpose(y);
end

